# !!!! WORK IN PROGRESSS !!!!
#
# 
# TEAM benchmark 23 with periodic excitation and eddy current via non linear A-Psi potentials
# https://www.compumag.org/jsite/team.html
#
# Nils Much, 25.04.2019
#
#
# !!!! WORK IN PROGRESSS !!!!


## -------------- Initialization ----------------------------------------------------------------------

# print initialization has started
print('\n--------Initialization---------')

# import libraries
import netgen.gui
from ngsolve import *
from netgen.csg import *
from netgen.meshing import Mesh as NgMesh
import numpy as np
import time
import sys
import nlinsolver

# unit system
# kg, s, mm, A

# initialize constants for this problem
# geometry
d1 = 3.048           #[mm]
d2 = 3.9624          #[mm]
d3 = 2.9972          #[mm]
l = 1.6              #[mm]
lc = 1.524           #[mm]
sc = 0.127           #[mm]
delta = 0.254        #[mm] # 
r = 10.0             #[mm]

# coil parameters
current = 0.1           #[A]
turns = 280             #[-]

# magnetic permeability of free space
mu0 = 1.257e-3          #[kg*mm/(s^2*A^2)]

# electric conductivity of iron
sigma_iron = 1e-2       #[A^2*s^3/(kg*mm^3)]

# frequency
frequency = 10000.0      #[1/s]
omega = 2*np.pi*frequency

# solver settings
mag_conditioning = 1e-3
#el_conditioning = 1e-3
tol_solver = 1e-10
preconditioner = "bddc" # "bddc", "multigrid"
max_steps_solver = 200

# set message level
ngsglobals.msg_level = 5


## -------------- Mesh Generation ----------------------------------------------------------------------

# print initialization has started
print('\n--------Mesh Generation--------')

# create geometry
geom = CSGeometry()
geom.SetBoundingBox(Pnt(-100,-100,-100), Pnt(100,100,100))
geom_iron = (Cylinder(Pnt(0,0,delta), Pnt(0,0,delta+l), d3/2) * Plane(Pnt(0,0,delta), Vec(0,0,-1)) * Plane(Pnt(0,0,delta+l), Vec(0,0,1))).bc("gamma_iron")
geom_coil = ((Cylinder(Pnt(0,0,-lc-sc), Pnt(0,0,-sc), d2/2)-Cylinder(Pnt(0,0,-lc-sc), Pnt(0,0,-sc), d1/2)) * Plane(Pnt(0,0,-lc-sc), Vec(0,0,-1)) * Plane(Pnt(0,0,-sc), Vec(0,0,1))).bc("gamma_coil")
geom_box = OrthoBrick(Pnt(-r,-r,-r), Pnt(r,r,r))
geom.Add(geom_iron.mat("iron").maxh(0.3)) # fine mesh needed in iron domain!
geom.Add(geom_coil.mat("coil").maxh(0.5))
geom.Add((geom_box-geom_iron-geom_coil).mat("air"))

# mesh geometry
mesh = Mesh(geom.GenerateMesh(maxh = 3))
mesh.Curve(3)

# print number of volume elements and number of border elements
print ("number of volume elements :", mesh.GetNE(VOL))
print ("number of border elements :", mesh.GetNE(BND))

# create H-Curl finite element space for solving magnet domain
fesCurl_cplx = HCurl(mesh, order = 3, nograds = True, complex = True) #
    # HCurl : creates a curl() differentialble function space
    # nograds=True : deletes all basis-functions which are gradients

# create H-1 finite element space in iron domain for solving eddy current
fes1_cplx_iron = H1(mesh, order = 3, complex = True) #, definedon = "iron"

# combine finte element spaces for mixed formulation
fesMixed = FESpace([fesCurl_cplx, fes1_cplx_iron])

# print number of degrees of freedom in mesh
print ("number of dof for A   :", fesCurl_cplx.ndof)
print ("number of dof for Psi :", fes1_cplx_iron.ndof)
print ("number of dof mixed   :", fesMixed.ndof)

# get trial function u_Curl and test function v_Curl
u_A, u_Psi = fesMixed.TrialFunction()
v_A, v_Psi = fesMixed.TestFunction()


## -------------- Allocate Material Parameters ----------------------------------------------------------------------

# define domains for postprocessing in Paraview
iron_domain = CoefficientFunction([1 if mat == "iron" else 0 for mat in mesh.GetMaterials()])
coil_domain = CoefficientFunction([1 if mat == "coil" else 0 for mat in mesh.GetMaterials()])

# set material parameters as coefficient functions
# inverse of magnetic permeability
mur = { "iron" : 100, "air" : 1, "coil" : 1 }
mu_inv = 1.0 / mu0 / CoefficientFunction([mur[mat] for mat in mesh.GetMaterials()]) #[s^2*A^2/(kg*mm)]

# current load on coil
tau = CoefficientFunction((-y, x, 0))
tau = 1.0/Norm(tau)*tau
curdens = current * turns  / (lc*(d2-d1)/2) # Amper turns / coil area
j_coil = CoefficientFunction([curdens*tau if mat == "coil" else (0,0,0) for mat in mesh.GetMaterials()]) #[A/(mm^2)]

# electric conductivity
sigma = sigma_iron * CoefficientFunction([1 if mat == "iron" else 0.0001 for mat in mesh.GetMaterials()]) #[kg*mm^3/(A^2*s^3)]


## -------------- Eddy Solver ----------------------------------------------------------------------

# print eddy solver have started
print('\n----------Eddy Solver----------')
    
# save starting time
start_time = time.time()

# initialize left hand side from finite element space
sysmat_Mixed = BilinearForm(fesMixed)
    
# apply weak form to left hand side
sysmat_Mixed += SymbolicBFI(
        mu_inv*curl(u_A)*curl(v_A) + 
        1j*omega*sigma*v_A*u_A + 
        1j*omega*sigma*v_A*grad(u_Psi)*u_Psi + 
        1j*omega*sigma*grad(v_Psi)*grad(u_Psi) + 
        1j*omega*sigma*grad(v_Psi)*u_A)

# initialize right hand side from fintite element space
rhs_Mixed = LinearForm(fesMixed)

# apply weak form to right hand side
rhs_Mixed += SymbolicLFI(v_A*j_coil)

# assemble right hand side
rhs_Mixed.Assemble()

# initialize grid function for solution
sol_Mixed = GridFunction(fesMixed, name = "Mixed")




# initialize residual vector
lhs = sol_Mixed.vec.CreateVector()
res = sol_Mixed.vec.CreateVector()

# initialize step vector
dsol = sol_Mixed.vec.CreateVector()

# enter solving loop
for it in range(10):
    print("Iteration {:3}  ".format(it), end="")
    
    # apply current solution to BilinearForm
    sysmat_Mixed.Apply(sol_Mixed.vec, lhs) # 
    
    # calculate residual vector
    res.data = rhs_Mixed.vec - lhs
    
    # create preconditioner
    pre = Preconditioner(sysmat_Mixed, "bddc") # "bddc", "multigrid"
    
    # assemble linearized system matrix
    sysmat_Mixed.AssembleLinearization(sol_Mixed.vec)
    
    # solve linear system of equations
    solvers.GMRes(A = sysmat_Mixed.mat, pre = pre.mat, b = res, 
               x = dsol, printrates = True, tol = 1e-10, maxsteps = 200)
    
    # update solution
    sol_Mixed.vec.data -= dsol
    
    # check convergence
    resval = sqrt(abs(InnerProduct(dsol, res)))
    print ("residual = ", resval)
    if resval < 1e-8:
        break

